//
//  maxi_embind.h
//  Maxi_Emscripten
//
//  Created by D J Clarke on 09/10/2015.
//  Copyright (c) 2015 D J Clarke. All rights reserved.
//

#ifndef Maxi_Emscripten_maxi_embind_h
#define Maxi_Emscripten_maxi_embind_h

#include <emscripten.h>
#include <bind.h>

class vectorTools {
public:
	//template <typename T>
	static void clearVectorDbl(vector<double>& vecIn) {
		vecIn.clear();
	}
};

using namespace emscripten;

EMSCRIPTEN_BINDINGS(my_module) {
	register_vector<int>("VectorInt");
	register_vector<double>("VectorDouble");
	//	register_vector<float>("VectorFloat");

	class_<vectorTools>("vectorTools")
	.constructor<>()
	.class_function("clearVectorDbl", &vectorTools::clearVectorDbl)
	;

	class_<maxiSettings>("maxiSettings")
	.constructor<>()
	.class_function("setup", &maxiSettings::setup)
	;
	
	
	// MAXI OSC
	class_<maxiOsc>("maxiOsc")
	.constructor<>()
	.function("sinewave", &maxiOsc::sinewave)
	.function("coswave", &maxiOsc::coswave)
	.function("phasor", select_overload<double(double)>(&maxiOsc::phasor))
	.function("phasor", select_overload<double(double, double, double)>(&maxiOsc::phasor))
	.function("saw", &maxiOsc::saw)
	.function("triangle", &maxiOsc::triangle)
	.function("square", &maxiOsc::square)
	.function("pulse", &maxiOsc::pulse)
	.function("noise", &maxiOsc::noise)
	.function("sinebuf", &maxiOsc::sinebuf)
	.function("sinebuf4", &maxiOsc::sinebuf4)
	.function("sawn", &maxiOsc::sawn)
	.function("rect", &maxiOsc::rect)
	.function("phaseReset", &maxiOsc::phaseReset)
	;
	
	// MAXI ENVELOPE
	class_<maxiEnvelope>("maxiEnvelope")
	.constructor<>()
	.function("line", &maxiEnvelope::line)
	//	.function("line", &maxiEnvelope::line, allow_raw_pointers()) // if using array version
	.function("trigger", &maxiEnvelope::trigger)
	.property("amplitude", &maxiEnvelope::GetAmplitude, &maxiEnvelope::SetAmplitude)
	;
	
	// MAXI DELAYLINE
	class_<maxiDelayline>("maxiDelayline")
	.constructor<>()
	.function("dl", select_overload<double(double, int, double)>(&maxiDelayline::dl))
	.function("dl", select_overload<double(double, int, double, int)>(&maxiDelayline::dl))
	;

	
	// MAXI FILTER
	class_<maxiFilter>("maxiFilter")
	.constructor<>()
	.function("lores", &maxiFilter::lores)
	.function("hires", &maxiFilter::hires)
	.function("bandpass", &maxiFilter::bandpass)
	.function("lopass", &maxiFilter::lopass)
	.function("hipass", &maxiFilter::hipass)
	;
	
	// MAXI MIX
	class_<maxiMix>("maxiMix")
	.constructor<>()
	.function("stereo", &maxiMix::stereo, allow_raw_pointers())
	.function("quad", &maxiMix::quad,  allow_raw_pointers())
	.function("ambisonic", &maxiMix::ambisonic, allow_raw_pointers())
	;
	
	// MAXI SAMPLE
	class_<maxiSample>("maxiSample")
	.constructor<>()
	.function("setSample", &maxiSample::setSample)
//	.function("getSummary", &maxiSample::getSummary)
	
	.function("playOnce", &maxiSample::playOnce)
	.function("play", select_overload<double()>(&maxiSample::play))
	.function("play", select_overload<double(double)>(&maxiSample::play))
	.function("play", select_overload<double(double, double, double)>(&maxiSample::play))
//	.function("play", select_overload<double(double, double, double, double&)>(&maxiSample::play))
	.function("play4", &maxiSample::play4)


	.function("trigger", &maxiSample::trigger)
	.function("clear", &maxiSample::clear)

	//	.function("load", &maxiSample::load)
//	.function("read", &maxiSample::read, allow_raw_pointers())
	;
	
	// MAXI MAP
	class_<maxiMap>("maxiMap")
	.constructor<>()
	.function("linlin", &maxiMap::linlin)
	.function("linexp", &maxiMap::linexp)
	.function("explin", &maxiMap::explin)
	.function("clamp", &maxiMap::clamp)
	;
	
	// MAXI DYN
	class_<maxiDyn>("maxiDyn")
	.constructor<>()
	.function("gate", &maxiDyn::gate)
	.function("compressor", &maxiDyn::compressor)
	;
	
	// MAXI ENV
	class_<maxiEnv>("maxiEnv")
	.constructor<>()
	.function("ar", &maxiEnv::ar)
	.function("adsr", &maxiEnv::adsr)
	;
	
	// CONVERT
	class_<convert>("convert")
	.constructor<>()
	.function("mtof", &convert::mtof)
	;
	
	// MAXI DISTORTION
	class_<maxiDistortion>("maxiDistortion")
	.constructor<>()
	.function("fastAtan", &maxiDistortion::fastatan)
	.function("atanDist", &maxiDistortion::atanDist)
	.function("fastAtanDist", &maxiDistortion::fastAtanDist)
	;
	
	// MAXI FLANGER
	class_<maxiFlanger>("maxiFlanger")
	.constructor<>()
	.function("flange", &maxiFlanger::flange)
	;
	
	// MAXI CHORUS
	class_<maxiChorus>("maxiChorus")
	.constructor<>()
	.function("chorus", &maxiChorus::chorus)
	;
	
	// MAXI ENVELOPE FOLLOWER
	class_<maxiEnvelopeFollower>("maxiEnvelopeFollower")
	.constructor<>()
	.function("setAttack", &maxiEnvelopeFollower::setAttack)
	.function("setRelease", &maxiEnvelopeFollower::setRelease)
	.function("play", &maxiEnvelopeFollower::play)
	.function("reset", &maxiEnvelopeFollower::reset)
	;
	
};
#endif
